<?php


namespace Schaffen\Fiscal\NfeDownloader\Services;


use Illuminate\Support\Facades\Lang;
use NFePHP\NFe\Exception\DocumentsException;
use Schaffen\Core\Company\Repositories\BranchRepository;
use Schaffen\Core\Company\Repositories\CompanyRepository;
use Schaffen\Core\Person\Repositories\PersonRepository;
use Schaffen\Fiscal\NfeDownloader\Interfaces\ISefazSearch;
use Schaffen\Core\Person\Models\Person;
use Schaffen\Core\Company\Models\Branch;
use NFePHP\NFe\Tools;
use Illuminate\Support\Facades\DB;
use NFePHP\Common\Certificate;
use NFePHP\NFe\Common\Standardize;
use Schaffen\Fiscal\NfeDownloader\Repositories\AddressRepository;
use Schaffen\Fiscal\NfeDownloader\Repositories\FileContentRepository;
use Schaffen\Fiscal\NfeDownloader\Repositories\FileRepository;
use Schaffen\Fiscal\NfeDownloader\Repositories\IssuerRepository;
use Schaffen\Fiscal\NfeDownloader\Repositories\NFeRepository;
use Schaffen\Fiscal\NfeDownloader\Repositories\NFeSearchRepository;

//use function React\Promise\all;

class SefazDownloader implements ISefazSearch
{

    const AMBIENTE_PRODUCAO = 1;
    const AMBIENTE_HOMOLOGACAO = 2;
//    const PATH_CERTIFICADOS_SEFAZ = "/var/www/html/certificados_sefaz/";
    const PATH_CERTIFICADOS_SEFAZ = "/var/www/certificados_sefaz/";

    private $emitent;
    private $filial;
    private $company;
    private $issuer;
    private $nfe_version;
    private $person_repository;
    private $nfe_search_repository;
    private $file_content_repository;
    private $file_repository;
    private $address_repository;
    private $nfe_repository;
    private $branch_repository;

    public function __construct(
        CompanyRepository $company_repository,
        IssuerRepository $issuer_repository,
        PersonRepository $person_repository,
        NFeSearchRepository $nfe_search_repository,
        FileContentRepository $file_content_repository,
        FileRepository $file_repository,
        AddressRepository $address_repository,
        NFeRepository $nfe_repository,
        BranchRepository $branch_repository
    )
    {
        $this->company = $company_repository;
        $this->issuer = $issuer_repository;
        $this->person_repository = $person_repository;
        $this->nfe_search_repository = $nfe_search_repository;
        $this->file_content_repository = $file_content_repository;
        $this->file_repository = $file_repository;
        $this->address_repository = $address_repository;
        $this->nfe_repository = $nfe_repository;
        $this->branch_repository = $branch_repository;
    }

    public function getClients()
    {
        // TODO: Implement getClient() method.

        $companies = $this->company->list(['emp_usa_download_nfe' => 1]);

        if (is_null($companies)) {
            throw new ModelNotFound(Lang::get('exceptions.exception.db_not_found'));
        }

        foreach ($companies as $company) {

            $this->setDatabase($company);

            $branches = $this->getBranches();
            foreach ($branches as $branch) {
                $issuer = $this->person_repository->find($branch->fil_fk_pes_codigo, ['address']);
//                dd($issuer);
                if (is_null($issuer)) {
                    continue;
                }
                $pfx = $this->getPfx($issuer);

                if (!$pfx)
                    continue;
                $configJson = $this->getConfigTools($branch, $issuer);
                try {

                    $tools = new Tools($configJson, Certificate::readPfx($pfx, $branch->fil_certificado_passwd));

                } catch (DocumentsException $e) {

                }


                $tools->model('55');
                $contingencia = $tools->contingency->deactivate();
                $tools->contingency->load($contingencia);

                DB::beginTransaction();

                $procLimit = 100;
                $count = 0;


                do {

                    //trecho de código responsável por não repetir a consulta no sefaz quando vem o 137 (quando nao encontra nenhum documento)
                    $last_search = $this->nfe_search_repository->getLastSearchFromBranch($branch->fil_codigo);
                    if ($last_search) {
                        if ($last_search->status == 137) {
                            $now = strtotime(date("Y-m-d H:i:s"));
                            $last_response = strtotime($last_search->cnx_data_hora_consulta);
                            $diff = $now - $last_response;
                            if ($diff / (60 * 60) >= 1)
                                break;
                        }
                        $nsu = $last_search->cnx_ultimo_nsu;
                    } else {
                        $nsu = 0;
                    }

                    $response = $tools->sefazDistDFe(
                        $nsu
                    );

                    $resp = new Standardize($response);
                    $resp = $resp->toStd();


                    extract(get_object_vars($resp));

                    $dhResp = date("Y-m-d H:i:s", strtotime($dhResp));

                    $search = $this->nfe_search_repository->storeOrUpdate(
                        [
                            'cnx_ultimo_nsu' => $ultNSU,
                            'cnx_data_hora_consulta' => $dhResp,
                            'cnx_status' => $cStat,
                            'cnx_fk_fil_codigo' => $branch->fil_codigo,
                            'cnx_xml' => $response
                        ]
                    );

                    if (is_array($loteDistDFeInt->docZip)) {
                        foreach ($loteDistDFeInt->docZip as $doc) {
                            $this->importFromSearch($doc, $search->cnx_codigo);
                            $count++;
                        }
                    } else {
                        $doc = $loteDistDFeInt->docZip;
                        $this->importFromSearch($doc, $search->cnx_codigo);

                        $count++;
                    }


                } while ($cStat == 138 && $count < $procLimit);
                DB::commit();

            }


        }

    }

    public function importFromSearch($document, $searchId)
    {
        if (is_null($document))
            return false;
        DB::beginTransaction();
        $file_c = $this->file_content_repository->storeOrUpdate(['cxm_conteudo' => $document]);
        $decoded_doc = gzdecode(base64_decode($document));
        $std = $this->getStd($decoded_doc);
        if ($std) {
            $array_nfe = $this->storeNFe($std);
            if (!$array_nfe) {
                DB::rollBack();
                return;
            }
            $ak = $array_nfe["nfi_chave_acesso"];
            $filename = "NFe${ak}.xml";
            $description = "XML importado automaticamente";
            $file = $this->file_repository->storeOrUpdate([
                'aim_descricao' => $description,
                'aim_fk_cxm_codigo' => $file_c->cxm_codigo,
                'aim_nome_arquivo' => $filename
            ]);

            $array_nfe = array_merge($array_nfe, [
                'nfi_fk_aim_codigo' => $file->aim_codigo,
                'nfi_fk_cnx_codigo' => $searchId,
                'nfi_data_hora_emissao' => $array_nfe['nfi_data_hora_emissao']
            ]);
            $this->nfe_repository->storeOrUpdate($array_nfe);
        }

        DB::commit();
    }

    public function storeNFe($nfe)
    {
        $inf = $nfe->infNFe;
        $emit = $inf->emit;


        $Id = $inf->attributes->Id;
        $ak = str_replace('NFe', '', $Id);
        $nfe_ak = $this->nfe_repository->first(
            [
                'nfi_chave_acesso' => $ak
            ]
        );

        if ($nfe_ak)
            return false;
        if (!empty($emit->enderEmit)) {
            $xLgr = !empty($emit->enderEmit->xLgr) ? trim($emit->enderEmit->xLgr) : null;
            $nro = !empty($emit->enderEmit->nro) ? trim($emit->enderEmit->nro) : null;
            $xCpl = !empty($emit->enderEmit->xCpl) ? trim($emit->enderEmit->xCpl) : null;
            $xBairro = !empty($emit->enderEmit->xBairro) ? trim($emit->enderEmit->xBairro) : null;
            $xMun = !empty($emit->enderEmit->xMun) ? trim($emit->enderEmit->xMun) : null;
            $UF = !empty($emit->enderEmit->UF) ? trim($emit->enderEmit->UF) : null;
            $CEP = !empty($emit->enderEmit->CEP) ? trim($emit->enderEmit->CEP) : null;
            $xPais = !empty($emit->enderEmit->xPais) ? trim($emit->enderEmit->xPais) : null;

            $address_array = array(
                'eem_logradouro' => $xLgr,
                'eem_numero' => $nro,
                'eem_complemento' => $xCpl,
                'eem_bairro' => $xBairro,
                'eem_cidade' => $xMun,
                'eem_estado' => $UF,
                'eem_cep' => $CEP,
                'eem_pais' => $xPais
            );
            $address = $this->address_repository->first($address_array);
            if (!$address) {
                $address = $this->address_repository->storeOrUpdate($address_array);
            }
        }

        if (!empty($emit->CNPJ)) {
            $cpf_cnpj = trim($emit->CNPJ);
        } elseif (!empty($emit->CPF)) {
            $cpf_cnpj = trim($emit->CPF);
        } else {
            $cpf_cnpj = null;
        }
        $xNome = !empty($emit->xNome) ? trim($emit->xNome) : null;
        $xFant = !empty($emit->xFant) ? trim($emit->xFant) : null;
        $IE = !empty($emit->IE) ? trim($emit->IE) : null;
        $IM = !empty($emit->IM) ? trim($emit->IM) : null;

        $issuer = $this->issuer->first(
            [
                'emi_cpf_cnpj' => $cpf_cnpj
            ]
        );
        if (!$issuer) {
            $issuer_array = array(
                'emi_cpf_cnpj' => $cpf_cnpj,
                'emi_razao_social' => $xNome,
                'emi_nome_fantasia' => $xFant,
                'emi_fk_eem_codigo' => $address->eem_codigo,
                'emi_registro_estadual' => $IE,
                'emi_registro_municipal' => $IM
            );
            $issuer = $this->issuer->storeOrUpdate($issuer_array);
        }

        $ide = $inf->ide;
        $dhEmi = trim($ide->dhEmi);
        $datetime = date("Y-m-d H:i:s", strtotime($dhEmi) + 3600);
        $val = $inf->total->ICMSTot->vNF;

        return [
            'nfi_chave_acesso' => $ak,
            'nfi_valor' => $val,
            'nfi_data_hora_emissao' => $datetime,
            'nfi_fk_emi_codigo' => $issuer->emi_codigo
        ];
    }

    public function getStd($xml)
    {
        $st = new Standardize($xml);
        $std = $st->toStd();

        switch ($st->whichIs($xml)) {
            case 'nfeProc':
                return $std->NFe;
            case 'NFe':
                return $std;
        }
    }

    public function getPfx(Person $issuer)
    {
        $certPath = rtrim(static::PATH_CERTIFICADOS_SEFAZ, DIRECTORY_SEPARATOR);
        $certPath .= DIRECTORY_SEPARATOR . $issuer->pes_identificador . ".pfx";

        if (!file_exists($certPath)) {
            return false;
        }

        $pfx = @file_get_contents($certPath);

        return $pfx;
    }

    public function getConfigTools(Branch $branch, Person $issuer)
    {

        // TODO: Implement getConfigTools() method.
        $configJson = json_encode([
            "atualizacao" => date('Y-m-d h:i:s'),
            "tpAmb" => $this->getAmbienteEmissao($branch),
            "razaosocial" => $this->getAmbienteEmissao($branch) == static::AMBIENTE_HOMOLOGACAO ?
                "NF-E EMITIDA EM AMBIENTE DE HOMOLOGACAO - SEM VALOR FISCAL" : $issuer->pes_razao_social,
            "siglaUF" => $issuer->address[0]->pee_uf,
            "cnpj" => $issuer->pes_identificador,
            "schemes" => "PL_009_V4",
            "versao" => '4.00',
            "tokenIBPT" => "",
            "CSC" => $this->getAmbienteEmissao($branch) == static::AMBIENTE_HOMOLOGACAO ?
                $branch->fil_csc_homologacao : $branch->fil_csc_producao,
            "CSCid" => $this->getAmbienteEmissao($branch) == static::AMBIENTE_HOMOLOGACAO ?
                $branch->fil_csc_token_homologacao : $branch->fil_csc_token_producao,
            "aProxyConf" => [
                "proxyIp" => "",
                "proxyPort" => "",
                "proxyUser" => "",
                "proxyPass" => ""
            ]
        ]);

        return $configJson;

    }


    public function getAmbienteEmissao(Branch $branch)
    {
        return $branch->fil_sefaz_ambiente;
    }


    public function getBranches()
    {
        return $this->branch_repository->list();
    }

    public function setDatabase($company)
    {
        config([
            'database.connections.dynamicdb' => [
                'driver' => 'pgsql',
                'host' => $company->emp_url,
                'port' => $company->emp_porta,
                'database' => $company->emp_banco,
                'username' => env('DB_USERNAME'),
                'password' => env('DB_PASSWORD'),
                'schema' => 'cadastro'
            ]
        ]);
    }

    public function getNotes()
    {


        // TODO: Implement getNotes() method.
    }


}
